
unit ASWSMClient;

interface

uses
  WinSock;

procedure NewPeer(aSocket: tSocket; const aBasePath, aBaseFile: string);

var
  RC: Integer;

implementation

uses
  Windows, WSocket, Classes, SysUtils, HomeTool, ASWSPlugin, ASWSMain, HTTPProt{by F.P.};

type
  tPeerThread = class(tThread)
  private
    Peer      : tWSocket;
    fSock     : tSocket;
    PeerBuf   : string;
    Connected : Boolean;
    HeadAvail : Boolean;
    DataStart : Integer;
    noReply   : Boolean;
    FileOK    : Boolean;
    ClientDone: Boolean;
    ReqFile   : string;
    BaseFile  : string;
    BasePath  : string;
    DataSent  : Boolean;
    PeerHost  : string;
    gfhExt    : string;
    gfhHandle : Integer;
    gfhResult : string;
    LogItem   : string;
    fMaxTicks : dWord;
    fTimeOut  : dWord;
    fRequest  : tStrings;
    fPartial  : Boolean;
    fFileStart: Integer;
    fFileEnd  : Integer; 
    procedure PSC(Sender: TObject; Error: Word);
    procedure PDA(Sender: TObject; Error: Word);
    procedure PDS(Sender: TObject; Error: Word);
    procedure PCS(Sender: TObject; OldState, NewState: TSocketState);
    //
    procedure ParseReq(const aValue: string);
    function  TryThis(aStr, aSubStr: string): Boolean;
    procedure Reply(ErCode: Integer; const aMsg{, aOtherFields}: string);
    procedure Wait;
    function  GetFileHdr(const aFileExt: string; aFile: tHandle): string;
    procedure DoGetFileHdr;
    procedure SendFile(aHandle: Integer; const aFileName: string);
    procedure SendRequestedFile;
    procedure IncRC;
    procedure DecRC;
    procedure AddLog(const anItem: string);
    procedure DoAddLog;
    procedure AddPeer;
    function  HTTPField(const FieldName: string): string;
  public
    constructor Create(aSock: tSocket; const aBasePath, aBaseFile: string);
    procedure Execute; override;
    destructor Destroy; override;
  end;

{ tPeerThread }

constructor tPeerThread.Create;
begin
  fSock := aSock;
  BaseFile := aBaseFile;
  BasePath := aBasePath;
  FreeOnTerminate := True;
  fMaxTicks := (Str2Int(F_ASWSMain.Edit13.Text, 40) * 1000);
  fRequest  := tStringList.Create;
  inherited Create(True);
end;

function NumOnly(const aValue: string): string;
var
  i: Integer;
begin
  Result := '';
  i := 1;
  while (i <= Length(aValue)) and (aValue[i] < '0') do Inc(i);
  while (i <= Length(aValue)) and (aValue[i] in ['0'..'9']) do begin
    Result := Result + aValue[i];
    Inc(i);
  end;
end;

procedure tPeerThread.Execute;
var
  lZ: Integer;
  lS: Integer;
  lP: Integer;
  lB: Boolean;
  PeerName: TSockAddrIn;
begin
  Peer := tWSocket.Create(nil);
  with Peer do try
    OnSessionAvailable := PSC;
    OnDataAvailAble    := PDA;
    OnDataSent	       := PDS;
    OnChangeState      := PCS;
    //
    Peer.Dup(fSock);
    Peer.MultiThreaded := True;
    PeerBuf   := '';
    Connected := (Peer.State = wsConnected);
    Synchronize(IncRc);
    NoReply := True;
    fTimeOut := GetTickCount;
    if Connected then begin
      Peer.GetPeerName(PeerName, Sizeof(PeerName));
      PeerHost   := IntToStr(ord(PeerName.sin_addr.S_un_b.s_b1)) + '.' + IntToStr(ord(PeerName.sin_addr.S_un_b.s_b2)) + '.' + IntToStr(ord(PeerName.sin_addr.S_un_b.s_b3)) + '.' + IntToStr(ord(PeerName.sin_addr.S_un_b.s_b4));
      repeat
	Sleep(100);
	Peer.ProcessMessages;
	if not HeadAvail then begin
	  lZ := 4;
	  lS := Pos(#13#10#13#10, PeerBuf);
	  if (lS > 0) then begin
	    lP := Pos('CONTENT-LENGTH:', UpperCase(PeerBuf));
	    DataStart := lS + lZ;
	    if (lP > 0) then lB := Length(PeerBuf) >= DataStart - 1 + Str2Int(NumOnly(Copy(PeerBuf, lP + 15, Length(PeerBuf))), 0)
			else lB := True;
	    if lB then begin
	      HeadAvail := True;
	      try
		ParseReq(Copy(PeerBuf, 1, lS - 1));
	      except
		// ignore all errors
	      end;
	    end;
	  end;
	end;
	if not ClientDone and (GetTickCount - fTimeOut > fMaxTicks) then Break;
      until ClientDone or not Connected;
    end;
  finally
    if Connected then Peer.Close;
    while Connected do begin
      Sleep(100);
      Peer.ProcessMessages;
    end;
    Synchronize(DecRc);
  end;
end;

destructor tPeerThread.Destroy;
begin
  Peer.Free;
  fRequest.Free;
  inherited;
end;

procedure tPeerThread.IncRC;
begin
  Inc(RC);
  F_ASWSMain.ClientsNum := RC;
end;

procedure tPeerThread.DecRC;
begin
  Dec(RC);
  F_ASWSMain.ClientsNum := RC;
end;

type
  tErrorCode = record
    Code: Integer;
    Desc: string;
  end;

const
  ErrorCodes: array[1..40] of tErrorCode = (
    (Code: 100; Desc: 'Continue'),
    (Code: 101; Desc: 'Switching Protocols'),
    (Code: 200; Desc: 'OK'),
    (Code: 201; Desc: 'Created'),
    (Code: 202; Desc: 'Accepted'),
    (Code: 203; Desc: 'Non-Authoritative Information'),
    (Code: 204; Desc: 'No Content'),
    (Code: 205; Desc: 'Reset Content'),
    (Code: 206; Desc: 'Partial Content'),
    (Code: 300; Desc: 'Multiple Choices'),
    (Code: 301; Desc: 'Moved Permanently'),
    (Code: 302; Desc: 'Found'),
    (Code: 303; Desc: 'See Other'),
    (Code: 304; Desc: 'Not Modified'),
    (Code: 305; Desc: 'Use Proxy'),
    (Code: 307; Desc: 'Temporary Redirect'),
    (Code: 400; Desc: 'Bad Request'),
    (Code: 401; Desc: 'Unauthorized'),
    (Code: 402; Desc: 'Payment Required'),
    (Code: 403; Desc: 'Forbidden'),
    (Code: 404; Desc: 'Not Found'),
    (Code: 405; Desc: 'Method Not Allowed'),
    (Code: 406; Desc: 'Not Acceptable'),
    (Code: 407; Desc: 'Proxy Authentication Required'),
    (Code: 408; Desc: 'Request Time-out'),
    (Code: 409; Desc: 'Conflict'),
    (Code: 410; Desc: 'Gone'),
    (Code: 411; Desc: 'Length Required'),
    (Code: 412; Desc: 'Precondition Failed'),
    (Code: 413; Desc: 'Request Entity Too Large'),
    (Code: 414; Desc: 'Request-URI Too Large'),
    (Code: 415; Desc: 'Unsupported Media Type'),
    (Code: 416; Desc: 'Requested range not satisfiable'),
    (Code: 417; Desc: 'Expectation Failed'),
    (Code: 500; Desc: 'Internal Server Error'),
    (Code: 501; Desc: 'Not Implemented'),
    (Code: 502; Desc: 'Bad Gateway'),
    (Code: 503; Desc: 'Service Unavailable'),
    (Code: 504; Desc: 'Gateway Time-out'),
    (Code: 505; Desc: 'HTTP Version not supported')
  );

function GetHTTPStat(aStatus: Integer): string;
var
  CodeStr: string;
	i: Integer;
begin
  CodeStr := 'Unknown status';
  for i := Low(ErrorCodes) to High(ErrorCodes) do
    if (ErrorCodes[i].Code = aStatus) then begin
      CodeStr := ErrorCodes[i].Desc;
      Break;
    end;
  Result := 'HTTP/1.0 ' + IntToStr(aStatus) + ' ' + CodeStr + #13#10;
end;

procedure tPeerThread.Reply(ErCode: Integer; const aMsg{, aOtherFields}: string);
var
  lS: string;
  lZ: string;
begin
  lZ := '<HTML><h2>' + aMsg + '</h2></HTML>';
  lS := GetHTTPStat(ErCode) +
	'Server: ' + f_ASWSMain.fServerStr + #13#10 +
	//aOtherFields +
	'Content-Type: text/html'#13#10 +
	'Content-Length: ' + IntToStr(Length(lZ)) + #13#10#13#10 +
	lZ;
  DataSent := False;
  if Connected then begin
    Peer.SendStr(lS);
    Wait;
  end;
  noReply := False;
end;

function toDef(const S: string): string;
begin
  if (Trim(S) = '') then Result := 'binary/data'
		    else Result := S;
end;

procedure tPeerThread.DoGetFileHdr;
var
  lLength  : Integer;
  lSubRange: string;
begin
  lLength   := FileSeek(gfhHandle, 0, 2);
  lSubRange := '';
  if fPartial then begin
    lSubRange := '/' + IntToStr(lLength);
    if (fFileEnd >= 0) then lLength := Min(fFileEnd, lLength - 1)
		       else Dec(lLength);
    lLength := lLength - fFileStart + 1;
    lSubRange := 'Content-Range: bytes ' + IntToStr(fFileStart) + '-' + IntToStr(fFileStart + lLength - 1) + lSubRange + #13#10;
  end;
  gfhResult := 'Server: ' + f_ASWSMain.fServerStr + #13#10 +
	       'Last-Modified: ' + RFC1123_Date(FileDateToDateTime(FileGetDate(gfhHandle))) + ' GMT'#13#10 +
	       'Content-Type: ' + toDef(F_ASWSMain.MimeExtension[gfhExt]) + #13#10 +
	       'Content-Length: ' + IntToStr(lLength) + #13#10 +
	       lSubRange +
	       #13#10;
end;

function tPeerThread.GetFileHdr(const aFileExt: string; aFile: tHandle): string;
begin
  gfhExt := aFileExt;
  gfhHandle := aFile;
  Synchronize(DoGetFileHdr);
  Result := gfhResult;
end;

procedure tPeerThread.SendFile(aHandle: Integer; const aFileName: string);
const
  BufSize = 4096;	// not too big -- many client will overload memory
var
  B: pointer;	// local buffer
  M: Integer;   // total size to send
  S: Integer;	// sended bytes
  T: Integer;	// to read
begin
  M := FileSeek(aHandle, 0, 2);
  if fPartial then begin
    if (fFileEnd >= 0) then M := Min(M - 1, fFileEnd)
		       else Dec(M);
    S := fFileStart;
  end
  else begin
    Dec(M);
    S := 0;
  end;
  if (M >= 0) then begin
    GetMem(B, BufSize);
    try
      FileSeek(aHandle, S, 0);
      repeat
	T := Min(M - S + 1, BufSize);
	if (T > 0) then begin
	  FileRead(aHandle, B^, T);
	  DataSent := False;
	  if Connected then begin
	    Peer.Send(B, T);
	    Wait;
	  end
	  else Break;
	  Inc(S, T);
	end;
      until (T < 1) or not Connected;
    finally
      FreeMem(B, BufSize);
    end;
  end;
end;

function tPeerThread.HTTPField(const FieldName: string): string;
var
  NormFieldName: string;
  i	       : Integer;
  FieldValue   : string;
begin
  Result := '';
  NormFieldName := UpperCase(Trim(FieldName));
  if (NormFieldName <> '') then begin
    i := 1;
    while (i < fRequest.Count) do begin
      FieldValue := fRequest[i];
      if (Pos(NormFieldName, UpperCase(Trim(FieldValue))) = 1) then begin
	Result := Copy(FieldValue, Pos(':', FieldValue) + 1, Length(FieldValue));
	Break;
      end
      else Inc(i);
    end;
  end;
end;

function tPeerThread.TryThis(aStr, aSubStr: string): Boolean;
var
  i  : Integer;
  lP : Integer;
  lA : string;
  lF : string;
  lS : string;
  lCI,
  lCO: string;
  lFH: tHandle;
  lFO: tHandle;
  lPg: Boolean;
  lOK: Boolean;
  PGFields: string;
  Status  : Integer;
begin
  Result := False;
  aStr   := Trim(aStr);
  lP := Pos(aSubStr, aStr);
  if (lP = 1) then begin
    Delete(aStr, 1, lP + Length(aSubStr));
    aStr := Trim(aStr);
    if (Pos('HTTP://', UpperCase(aStr)) = 1) then begin
      i := 8;
      {$b-}
      while (i < Length(aStr)) and (aStr[i] <> '/') do Inc(i);
      Delete(aStr, 1, i - 1);
    end;
    lP := Pos('HTTP/1.', aStr);
    if (lP = Length(aStr) - 7) then ReqFile := Trim(Copy(aStr, 1, Pred(lP)))
			       else ReqFile := '';
    if (ReqFile = '/') then ReqFile := BaseFile;
    if (aSubStr = 'POST') then ReqFile := ReqFile + '?' + Copy(PeerBuf, DataStart, Length(PeerBuf));
    if (aSubStr = 'PUT') then begin
      lS := f_ASWSMain.Edit14.Text + ReqFile;
      if FileExists(lS) then Reply(406, 'File already exists')
      else begin
	lFH := FileCreate(lS);
	if (lFH > 0) then try
	  FileWrite(lFH, PeerBuf[DataStart], Length(PeerBuf) - DataStart + 1);
	finally
	  FileClose(lFH);
	  Reply(200, 'File saved');
	end
	else Reply(406, 'Cannot create such a file: ' + ReqFile);
      end;
      Exit;
    end;
    Result := (ReqFile <> '');
    if Result then begin
      lP := Pos('?', ReqFile);
      if (lP > 0) then begin
	lF  := Copy(ReqFile, 2, lP - 2);
	lA  := Copy(ReqFile, lP + 1, Length(ReqFile));
	lPg := (FindPlugin(lF) >= 0);
	lFH := 0;
	lFO := 0;
	if lPg or (lFH <> 0) then
	  try
	    if not lPg then begin
	      lOK := False;
	    end
	    else begin
	      PGFields := '';
	      Status := 200;
	      lOK := (QueryPlugin2(lF, PeerHost, {Copy(}PeerBuf{, 1, DataStart)}, lA, lCO, PGFields, Status) >= 0);
	      lA := F_ASWSMain.Header + lCO + F_ASWSMain.Footer;
	      if lOK then begin
		if (Pos('CONTENT-TYPE: ', UpperCase(PGFields)) < 1) then PGFields := PGFields + 'Content-type: text/html'#13#10;
		lS := GetHTTPStat(Status) + PGFields + 'Content-Length: ' + IntToStr(Length(lA)) + #13#10'Content:' +  #13#10#13#10 + lA;
		DataSent := False;
		if Connected then begin
		  Peer.SendStr(lS);
		  Wait;
		end;
	      end;
	    end;
	    if lOK then begin
	      noReply := False;
	      if not lPg then begin
		lS := GetHTTPStat(200) + GetFileHdr('HTM', lFH);
		DataSent := False;
		if Connected then begin
		  Peer.SendStr(lS);
		  Wait;
		  SendFile(lFH, lCO);
		end
	      end;
	    end
	    else begin
	      Reply(405, 'Cannot run plugin (' + ReqFile + ')');
	      Result := False;
	    end;
	  finally
	    if not lPg then begin
	      FileClose(lFH);
	      FileClose(lFO);
	      DeleteFile(lCO);
	      DeleteFile(lCI);
	    end;
	  end
	else begin
	  Reply(500, 'Cannot create temporary file. CGI fail.');
	  Result := False;
	end;
      end
      else begin
	FileOK := FileExists(BasePath + ReqFile);
	if FileOK then begin
	  fFileStart := 0;
	  fFileEnd   := -1;
	  fPartial   := False;
	  lS := HTTPField('Range');
	  if (Pos('BYTES', UpperCase(Trim(lS))) = 1) then begin
	    i := 1;
	    {$b-}
	    while (i < Length(lS)) and (lS[i] <> '=') do Inc(i);
	    Delete(lS, 1, i);
	    i := 1;
	    while (i < Length(lS)) and (lS[i] <> '-') do Inc(i);
	    lP := Str2Int(Copy(lS, 1, i - 1), -1);
	    if (lP >= 0) then begin
	      fPartial   := True;
	      fFileStart := lP;
	      Delete(lS, 1, i);
	      fFileEnd := Str2Int(lS, -1);
	      if (fFileEnd > 0) and (fFileEnd < fFileStart) then begin
		fFileStart := 0;
		fFileEnd   := -1;
		fPartial   := False;
	      end;
	    end;
	  end;
	end;
      end;
    end;
  end;
end;

procedure tPeerThread.SendRequestedFile;
var
  fH: Integer;
  lS: string;
  lCode: Integer;
begin
  fH := FileOpen(BasePath + ReqFile, fmOpenRead + fmShareDenyWrite);
  if (fH <> -1) then try
    DataSent := False;
    if fPartial then lCode := 206
		else lCode := 200;
    lS := GetHTTPStat(lCode) + GetFileHdr(Copy(ExtractFileExt(ReqFile), 2, MAX_PATH), fH);
    if Connected then begin
      Peer.SendStr(lS);
      Wait;
      SendFile(fH, ReqFile);
    end
  finally
    FileClose(fH);
  end
  else Reply(404, 'Unable to open this file for read');
end;

procedure tPeerThread.ParseReq(const aValue: string);
var
  lVs: string;
  isF: Boolean;
  //Proto, User, Pass, Host, Port, Path : String;
begin
  try
    fRequest.Text := aValue;
    if (fRequest.Count > 0) then begin
      lVs := fRequest[0];
      AddLog(lVs);
      if f_ASWSMain.CB1.Checked then Synchronize(AddPeer);
      //ParseURL(lVs, Proto, User, Pass, Host, Port, Path);
      //if (Proto + User + Pass + Host + Port + Path = '') then Exit;
      IsF := TryThis(lVs, 'GET');
      if not IsF then IsF := TryThis(lVs, 'POST');
      if not isF then IsF := TryThis(lVs, 'PUT');
      if not IsF then
	if noReply then Reply(501, 'Request not supported')
	else
      else
	if not FileOK and noReply then Reply(404, '<h2>File not really found</h2>')
	else
	  if noReply then SendRequestedFile;
    end
    else Reply(400, 'You doesn''t say anything! Try again');
    Sleep(200);
    Peer.ProcessMessages;
  finally
    ClientDone := True;
  end;
end;

procedure tPeerThread.PSC(Sender: TObject; Error: Word);
begin
  Connected := False;
end;

procedure tPeerThread.PDA(Sender: TObject; Error: Word);
var
  BufSize: Integer;
  lSz: Integer;
  lOL: Integer;
begin
  fTimeOut := GetTickCount;	// reset timeout
  BufSize := Peer.RcvdCount;
  lOL := Length(PeerBuf);
  SetLength(PeerBuf, lOL + BufSize);
  lSz := Peer.Receive(@PeerBuf[lOL + 1], BufSize);
  SetLength(PeerBuf, lOL + lSz);
end;

procedure tPeerThread.PDS(Sender: TObject; Error: Word);
begin
  DataSent := True;
end;

procedure tPeerThread.PCS(Sender: TObject; OldState, NewState: TSocketState);
begin
  Connected := (NewState = wsConnected);
end;

procedure tPeerThread.Wait;
var
  lTimeOut: dWord;
begin
  lTimeOut := GetTickCount;
  repeat
    Sleep(10);
    Peer.ProcessMessages;
  until DataSent or not Connected or (GetTickCount - lTImeOut > fMaxTicks);
  Sleep(20);
  Peer.ProcessMessages;
end;

procedure tPeerThread.AddLog(const anItem: string);
begin
  LogItem := anItem;
  Synchronize(DoAddLog);
end;

procedure tPeerThread.DoAddLog;
begin
  with F_ASWSMain do
    if CB5.Checked then
      AddTextToFile(Edit12.Text, DateTimeToStr(Now) + #9 + PeerHost + #9 + LogItem, True);
end;

procedure tPeerThread.AddPeer;
begin
  f_ASWSMain.M3.Text := f_ASWSMain.M3.Text + #13#10'===> From ' + PeerHost + ' at ' + TimeToStr(Time) + ' ======='#13#10 + fRequest.Text;
end;

//

procedure NewPeer(aSocket: tSocket; const aBasePath, aBaseFile: string);
begin
  with tPeerThread.Create(aSocket, aBasePath, aBaseFile) do Resume;
end;

end.

